/*****************************************************************************
 * $Id: parser.l,v 1.70 2003/02/10 20:54:36 jimb Exp $
 * Author		: Jim Brooks <hypersrc@jimbrooks.org>
 * Description	: Lexes function definitions and function calls in C/C++ code.
 *****************************************************************************/

%{

#define PARSER_L

#if 0
#define DEBUG
#endif

#undef yywrap
#define YY_NEVER_INTERACTIVE	1
#define YY_NO_UNPUT			/* shutup warning: `yyunput' defined but not used */

/*
 * Bits for discarding tokens.
 */
#define DISCARD_C_COMMENT				1
#define DISCARD_CPLUSPLUS_COMMENT		2
#define DISCARD_QUOTED					4
#define DISCARD_DEFINE					8
#define DISCARD_C_COMMENT_OR_QUOTED		(DISCARD_C_COMMENT | DISCARD_QUOTED)

#include "y.tab.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include "parser.h"
#include "str_heap.h"
#include "funcgraph.h"

extern STR_HEAP strHeap;	/* created at parser.y */
extern int yylex();
extern int parserMode;
extern int parsingFuncDef;
extern int parsedLeftParOfFuncDef;
extern int parsingCallParms;
extern int betweenParmsAndBody;
#if 1
extern char parserInputName[];
#else
extern char* parserInputName; /* --- this caused a segfault, why? --- */
#endif

/*
 * Vars which must be reset by LexReset().
 */
       int lineLex				= 1;
static int lineControlStmt		= -1;
static int discard				= 0;		/* this is a bit-field */
static int parentheses			= 0;
static int braces				= 0;
static int prevToken			= 0;		/* possibly changed by LexToken() as another token */
static int cast					= 0;
static int mismatchBraces		= 0;
static int mismatchParentheses	= 0;

/*
 * Other variables.
 */
int lexedIdLen;

/*
 * For non-lexical internal errors caused bugs, insufficent memory, bad pointers, etc.
 */
#define LEX_INTERNAL_ERROR_MSG( format, args... ) \
   fprintf( stderr, "lex INTERNAL ERROR: " )    , \
   fprintf( stderr, format, ## args )           , \
   fprintf( stderr, "\n" )

/*
 * For lexical errors present in the input being lexed.
 */
void LEXICAL_ERROR_MSG( char* str )
{
#ifndef LEX_YACC_QUIET
   if ( parserInputName[0] )
      fprintf( stderr, "%s:[%d]", parserInputName, lineLex );
   else
      fprintf( stderr, "[%d]", lineLex );

   fprintf( stderr, " lexical error" );
   if ( str[0] ) fprintf( stderr, ": %s", str ); 

   fprintf( stderr, "\n" );
#endif
}

/*
 * What to do if an internal error happens.
 */
#define LEX_ABORT() exit(1);

%}


/*===========================================================================*/

/*****************************************************************************
 * - regex notes:  "." means any char except newline.
 * - PreprocessCPlusPlusJava() overwrites start/end brace chars with
 *   ASCII 2,3 (\2,\3) so that methods defined within classes can be parsed.
 *****************************************************************************/

%%

						/* Filter-out parentheses within quotes. */
\".*\(.*\).*\"			{ return LexToken( TOKEN_TEXT ); }

						/* Pattern that matches either a stmt with a func call
						 * or a func prototype (both are very similar).
                         * Subpattern matches the (*foo) of a func ptr proto:
                         * (\([ \t]*\*[a-zA-Z_]+[ \t]*\))?
						 */
[a-zA-Z0-9_]+[ \t]*(\([ \t]*\*[a-zA-Z_]+[ \t]*\))?"("[^)(}{\2\3;/]*")"[^)(}{\2\3;/\"]*;		{
                          /* Count any newlines that may have been scooped. */
                          lineLex += OccurrencesInString( yytext, '\n' );
						  if (    parsingFuncDef  &&  !discard
                               && *(u_int32_t*)yytext != ('i'|'f'<<8|' '<<16|'('<<24) )
                          {
						     /* This is a stmt consisting of a func call. */
                             return LexToken( TOKEN_FUNC_CALL );
                          }
                          else
                          {
						     /* This is a function prototype. */
                             return LexToken( TOKEN_TEXT );
                          }
						}

\'\"\'		|			/* Ignore a double-quote enclosed by single-quotes or escaped. */
\\\"					{ return LexToken( TOKEN_TEXT ); }

"#endif"				{ /* Try to recover from an apparently mismatched parenthesis/brace */
                          /* that straddles '#'if/'#'else/'#'endif but would appear matched by cpp. */
                          LexErrorRecover();
                          return LexToken( TOKEN_TEXT );
                        }

\"[ ]*\#if[ ]*\"	|	/* Ignore keywords within quotes (strings) */
\"[ ]*\#elif[ ]*\"	|
\"[ ]*\#pragma[ ]*\"	{ return LexToken( TOKEN_TEXT ); }

\#if.*$					{ ; /* Ignore, eg "#if (ARCH==X86) && (SMP)" would confuse parser. */ }
\#elif.*$				{ ; }
\#pragma.*$				{ ; }

\#define				{ /* Discard tokens until last line of #define is encountered		*/
                          /* which is the next newline with a continuation char '\'.		*/
						  /* #define macros can contain parentheses which confuse parser.	*/
                          if ( !discard ) discard |= DISCARD_DEFINE; }
\\[ ]*\n				{ /* To discard tokens during a #define, don't let a newline	*/
                          /* with a continuation char '\' reach the below lex rule that	*/
                          /* clears DISCARD_DEFINE.										*/
                          lineLex++; }

if[ \t]*\([^()]*\)		|
for[ \t]*\([^()]*\)		|
while[ \t]*\([^()]*\)	|
switch[ \t]*\([^()]*\)	{ /* For speed, skip parsing those control statements. */
                          /* But pass to parser if parentheses contain nested parentheses */
                          /* which could contain function calls. */
                          return LexToken( TOKEN_TEXT ); }

sizeof\([^\)]*\)		{ /* Ignore "sizeof()". */
                          /* Don't greedily lex all right parentheses, eg: "if (x==sizeof(x))". */
                          return LexToken( TOKEN_TEXT ); }

virtual.*\(.*\).*;		{ /* Ignore C++ virtual prototypes. */
                          return LexToken( TOKEN_TEXT ); }

\/\/					{ /* Ignore "//" if it's enclosed by a C comment or quotes. */
                          if ( (discard & DISCARD_C_COMMENT_OR_QUOTED) == 0 )
                             discard |= DISCARD_CPLUSPLUS_COMMENT;
						}

"/*"					{ discard |=  DISCARD_C_COMMENT; }

"*/"					{ discard &= ~DISCARD_C_COMMENT; }

\"						{ /* Ignore quotes enclosed in comments. */ 
                          if ( ! (discard & (DISCARD_C_COMMENT|DISCARD_CPLUSPLUS_COMMENT)) )
                          {
                             discard ^= DISCARD_QUOTED;
                             if ( ! (discard & DISCARD_QUOTED) ) return TOKEN_TEXT;
                          }
                        }

                        /* '~' followed by id chars could a C++ deconstructor name */
[a-zA-Z_~][a-zA-Z_0-9]+	{ lexedIdLen = yyleng; return LexToken( TOKEN_ID ); }
[a-zA-Z_]				{ lexedIdLen = yyleng; return LexToken( TOKEN_ID ); }

\}[a-zA-Z0-9_ ]*;		{ /* Filter-out a right brace that is part of the end	*/
                          /* of a struct/union (including a typedef-ed struct).	*/
                          /* It would resemble:  "};"  or  "} type;"			*/
                          return LexToken( TOKEN_IGNORE ); }

						/* Only lex a left brace if it belongs to a func def. */
						/* Only lex a right brace if on start of line and it belongs to a func def. */
"{"						{ return LexToken( parsedLeftParOfFuncDef ? yytext[0] : TOKEN_TEXT ); }
^\}						{ int res; /* End of a func def is a good point to recover lex from any error. */
                          res  = LexToken( parsedLeftParOfFuncDef ? yytext[0] : TOKEN_TEXT );
                          LexErrorRecover();
                          return res; }

						/* PreprocessCPlusPlusJava() overwrites start/end brace chars with */
						/* ASCII 2,3 so that methods defined within classes can be parsed. */
\2						{ return LexToken( '{' ); }
\3						{ return LexToken( '}' ); }


[\(\)\,]				{ return LexToken( yytext[0] ); }

[ \t\r]+				{ ; }

\n						{ lineLex++;
                          if ( discard & DISCARD_CPLUSPLUS_COMMENT ) discard &= ~DISCARD_CPLUSPLUS_COMMENT;
                          if ( discard & DISCARD_DEFINE            ) discard &= ~DISCARD_DEFINE; }

.						{ return LexToken( TOKEN_TEXT ); }

%%


/*----------------------------------------------------------------------------
 * public functions
 *---------------------------------------------------------------------------*/


/*****************************************************************************
 * Tell lex to use a buffer.
 *
 * Returns:		Buffer's handle (really a pointer to YY_BUFFER_STATE).
 *****************************************************************************/
void*
LexBufferUse( char* buf )
{
   LexReset();
   yyrestart( NULL );
   return (void*) yy_scan_string( buf );
}

/*****************************************************************************
 * Discard buffer.
 *
 * Parms   : hnd
 *           Pass the handle returned by LexInputUse().
 *****************************************************************************/
void
LexBufferDiscard( void* hnd )
{
#ifdef HYPERSRC_PEDANTIC
   if ( hnd == NULL ) { LEX_INTERNAL_ERROR_MSG( "LexBufferDiscard() was passed NULL" ); return; }
#endif
   yy_delete_buffer( (YY_BUFFER_STATE) hnd );
}


/*----------------------------------------------------------------------------
 * internal functions
 *---------------------------------------------------------------------------*/


#define RETURN_TOKEN_ERROR() \
{ \
   if ( parserMode == PARSER_MODE_HYPERSRC ) \
      return TOKEN_ERROR; \
   else \
      LEX_ABORT(); \
}

#ifdef DEBUG
#define PRINT_TOKEN( lextoken ) \
{ \
   if ( 1 || ! discard ) \
   { \
      printf( "token = %d ", lextoken ); \
      if ( lextoken < 100 ) printf( " " ); \
      printf( " '%s' \n", yytext ); \
   } \
}
#else
#define PRINT_TOKEN( lextoken )
#endif /* DEBUG */

#define RETURN_TOKEN( lextoken ) \
{ \
   PRINT_TOKEN( lextoken ); \
   return prevToken = lextoken; \
} 

/*****************************************************************************
 * Process a lexeme/token.
 *
 * - record lexeme in a heap of strings
 * - assist parser by essentially preprocesing some tokens
 * - return a token to parser
 *
 * Parms   : yytext, yyleng (GLOBAL)
 *           Ptr,len of lexeme.
 *
 *           token
 *           Token to be returned to yacc.
 *
 * Statistics gathered by browsing /usr/src/sys/net* (FreeBSD):
 *
 * count:	token:
 * 9409		123 ?
 * 2070		125 ?
 * 615585	257 TOKEN_ID
 * 14163	258 TOKEN_FUNC_CALL
 * 448257	259 TOKEN_TEXT
 * 1479		261 TOKEN_IGNORE
 * 38784	40
 * 52859	41
 * 45565	44
 *****************************************************************************/
static int
LexToken( int token )
{
#ifdef HYPERSRC_DEBUG
   if (    (yyleng != strlen(yytext))
        && (yyleng != 1  &&  yytext[0] != '\0' ) ) /* strlen() won't count null char */
   {
      LEX_INTERNAL_ERROR_MSG( "%d %d %d '%s' %d",
                              yyleng, strlen(yytext), token, yytext, yytext[0] );
      LEX_ABORT();
   }
#endif

  /*
   * When a token is returned to yacc, yacc expects yylval to be assigned by lex.
   * Append this lexeme to a heap of strings.
   */
   yylval.str = StrHeapAppendFast( &strHeap, yytext, strlen(yytext) );

  /*
   * Was text lexed?
   */
   if ( token == TOKEN_TEXT )
      { RETURN_TOKEN( TOKEN_TEXT ); }

  /*
   * Discard tokens within comments or quotes.
   */
   if ( discard ) { RETURN_TOKEN( TOKEN_TEXT ); }

  /*
   * Was a func call lexed?
   */
   if ( token == TOKEN_FUNC_CALL )
      { RETURN_TOKEN( TOKEN_FUNC_CALL ); }

  /*
   * Was an identifier lexed?
   */
   if ( token == TOKEN_ID )
   {
     /*
      * If the token is a TOKEN_ID but is part of a cast,
      * then change the token to TOKEN_TEXT to avoid confusing the parser.
      */
      if ( !cast )
      {
         RETURN_TOKEN( TOKEN_ID );
      }
      else
      {
         DebugMsg( "changing TOKEN_ID to TOKEN_TEXT" );
         RETURN_TOKEN( TOKEN_TEXT );
      }
   }

  /*
   * Constructing a yacc rule to parse "int a, b, c;" is difficult.
   * This is an ugly alternative.
   */
   if ( token == ','  &&  parentheses == 0 )
      { RETURN_TOKEN( TOKEN_TEXT ); }

  /*
   * To simplify the parser rules, ignore nested parentheses and type casts.
   */
   if ( token == '(' )
   {
     /*
      * So that the parser can handle K&R parameter types that contain parentheses,
      * ignore parentheses between ')' and '{' in beginning of a func def.
      */
      if ( betweenParmsAndBody )
         { RETURN_TOKEN( TOKEN_TEXT ); }

      ++parentheses;
      if ( parentheses < 0 )
      {
         mismatchParentheses = TRUE;
         LEXICAL_ERROR_MSG( "parenthesis mismatch" );
         RETURN_TOKEN_ERROR();
      }

     /*
      * Function calls are assumed to begin with an identifier.
      * If this token is '(' but the previous token wasn't an identifier,
      * then assume it is the beginning of a type cast.
      * Ignore a cast because its syntactic similiarity to a function call
      * would confuse the function call yacc rules.
      */
      if ( prevToken != TOKEN_ID  &&  !cast )
      {
         if ( !parsingFuncDef  &&  lineLex != lineControlStmt )
         {
            cast = parentheses;		/* remember the parenthesis level */
         }
         RETURN_TOKEN( TOKEN_TEXT );
      }
#if 1
      RETURN_TOKEN( '(' );
#else /* keep this code in case it needs to be reinstated */
      if ( parentheses == 1  ||  parsingCallParms )
      { RETURN_TOKEN( '(' ); }		/* probably belongs to a function def/call */
      else
      { RETURN_TOKEN( TOKEN_TEXT ); }		/* ignore this nested parenthesis */
#endif
   }
   else if ( token == ')' )
   {
     /*
      * So that the parser can handle K&R parameter types that contain parentheses,
      * ignore parentheses between ')' and '{' in beginning of a func def.
      */
      if ( betweenParmsAndBody )
         { RETURN_TOKEN( TOKEN_TEXT ); }

      --parentheses;
      if ( parentheses < 0 )
      {
         mismatchParentheses = TRUE;
         LEXICAL_ERROR_MSG( "parenthesis mismatch" );
         RETURN_TOKEN_ERROR();
      }

     /*
      * If a '(' of a cast was parsed, does this ')' match it?
      */
      if ( cast  &&  (parentheses+1 == cast) )
         { cast = 0; RETURN_TOKEN( TOKEN_TEXT ); }
#if 1
      RETURN_TOKEN( ')' );
#else /* keep this code in case it needs to be reinstated */
      if ( parentheses == 0  ||  parsingCallParms )
      { RETURN_TOKEN( ')' ); }
      else
      { RETURN_TOKEN( TOKEN_TEXT ); }
#endif
   }

  /*
   * To simplify the parser rules, ignore nested braces.
   */
   if ( token == '{' )
   {
      ++braces;
      if ( braces < 0 )
      {
         mismatchBraces = TRUE;
         LEXICAL_ERROR_MSG( "curly brace '{' mismatch" );
         RETURN_TOKEN_ERROR();
      }

      if ( braces == 1 )
      { RETURN_TOKEN( '{' ); }
      else
#if 0
      { RETURN_TOKEN( TOKEN_TEXT ); }
#else
      /*
       * A func def might have more than one beginning '{' if
       * a #ifdef conditionally compiles different parameters types.
       * Try to ignore a superfluous '{'.
       */
      { braces = 1; RETURN_TOKEN( TOKEN_TEXT ); }
#endif
   }
   else if ( token == '}' )
   {
      --braces;
      if ( braces < 0 )
      {
         mismatchBraces = TRUE;
         LEXICAL_ERROR_MSG( "curly brace '}' mismatch" );
         RETURN_TOKEN_ERROR();
      }

      if ( braces == 0 )
      { RETURN_TOKEN( '}' ) }
      else
      { RETURN_TOKEN( TOKEN_TEXT ) }
   }

  /*
   * Return numeric value of token to yacc.
   */
   RETURN_TOKEN( token );
}

/*****************************************************************************
 * Must be called before lexing a new buffer.
 *****************************************************************************/
static void
LexReset( void )
{
   lineLex				= 1;
   lineControlStmt      = -1;
   discard				= 0;
   parentheses			= 0;
   braces				= 0;
   prevToken			= 0;
   cast					= 0;
   mismatchBraces		= 0;
   mismatchParentheses	= 0;
}

/*****************************************************************************
 * Recover if any error happened (can be called if there was no error).
 *****************************************************************************/
static void
LexErrorRecover( void )
{
   if ( mismatchParentheses ) { cast = 0;
                                mismatchParentheses = parentheses = prevToken = 0; }
   if ( mismatchBraces      ) { mismatchBraces      = braces      = prevToken = 0; }
}

/*****************************************************************************
 * What to do if end of input is reached.
 *****************************************************************************/
int
yywrap()
{
   if ( parserMode == PARSER_MODE_HYPERSRC )
   {
#if 1
      return 1;
#else
      yyterminate();
#endif
   }

   exit(0);
   return 0; /* shutup compiler */
}
