/*****************************************************************************
 * $Id: parser.y,v 1.35 2003/02/10 21:06:15 jimb Exp $
 * Author		: Jim Brooks <hypersrc@jimbrooks.org>
 * Description	: Parses function definitions and function calls in C/C++ code.
 *                This parser isn't perfect.  It might miss a few function calls,
 *                and misjudge some lines as parse errors where a compiler wouldn't.
 *
 * Notes        : Note that function calls within a function definition are
 *                reported before the function definition itself.
 *                The reason is the token for '}' is parsed last (last line #).
 *                However, before '}' is parsed, the id (name) of the func def
 *                has been parsed and recorded.
 *
 *                Eg:
 *                LoadFile{} calls fclose()  line # 67 
 *                LoadFile{} calls g_free()  line # 82 
 *                LoadFile{} calls Popup()  line # 83 
 *                LoadFile{} body resides at lines 32..85 
 *****************************************************************************/

/*
 * token struct ($<str>$).
 */
%union
{
   char* str;
}

/*
 * tokens
 */
%nonassoc TOKEN_ID
%nonassoc TOKEN_FUNC_CALL
%nonassoc TOKEN_TEXT
%nonassoc TOKEN_ERROR
%nonassoc TOKEN_IGNORE


/*===========================================================================*/


%{

#define PARSER_Y

#if 0
#define DEBUG
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include "parser.h"
#include "str_heap.h"
#include "funcgraph.h"

int yyerror( );

enum
{
   RESET_FULL,		/* full reset */
   RESET_STMT,		/* reset parser after a statement is sucessfully parsed */
   RESET_ERROR		/* reset parser after a parse error occurs */
};

STR_HEAP strHeap = { NULL, 0, NULL };
static char funcDefId[MAX_ID_LEN];
static char funcCallId[MAX_ID_LEN];
static char parsedId[MAX_ID_LEN];
static int  parsedIdLen = 0;
       char parserInputName[MAX_ID_LEN];

int parserMode = PARSER_MODE_PRINTF;

/*
 * Vars that should be initialized once every time ParserCreate() is called.
 */
static int initialized			= FALSE;
static int lineLeftParenthesis	= 0;
static int lineLeftBrace		= 0;
static int lineRightBrace		= 0;

/*
 * Vars that should be reset whenever ParserCreate() is called
 * and/or a parse error occurs.
 */
int parsingFuncDef				= FALSE;
int parsedLeftParOfFuncDef		= FALSE;
int parsingCallParms			= FALSE;
int betweenParmsAndBody			= FALSE;

extern char* yytext;
extern int yylex();
extern int lineLex;
extern int lexedIdLen;

/*****************************************************************************
 * For non-parse internal errors caused bugs, insufficent memory, bad pointers, etc.
 *****************************************************************************/
#define YACC_INTERNAL_ERROR_MSG( format, args... ) \
   fprintf( stderr, "yacc INTERNAL ERROR: " )    , \
   fprintf( stderr, format, ## args )            , \
   fprintf( stderr, "\n" )

/*****************************************************************************
 * For parse errors present in the input being parsed.
 *****************************************************************************/
void PARSE_ERROR_MSG( char* str )
{
#ifdef LEX_YACC_QUIET
   return;
#endif

   if ( parserInputName[0] )
      fprintf( stderr, "%s:[%d]", parserInputName, lineLex );
   else
      fprintf( stderr, "[%d]", lineLex );

   fprintf( stderr, " parse error" );

   if ( str[0] ) fprintf( stderr, ": %s", str ); 

   fprintf( stderr, "\n" );
}

/*****************************************************************************
 * What to do if an internal error happens.
 *****************************************************************************/
#define YACC_ABORT() exit(1);

/*****************************************************************************
 * Record a parsed identifier into an ASCIIZ string.
 * Identifier will be truncated if it exceeds MAX_ID_LEN.
 *
 * Parms : dest
 *         This string must be at least MAX_ID_LEN long.
 *
 *         src
 *         Must point to the first char of a presumed identifier,
 *         but doesn't matter if extra chars past the identifier were lexed.
 *         (Lex scoops more than an id for TOKEN_FUNC_CALL.)
 *
 *         lenSrc
 *         Must be the exact length of the identifier portion.
 *****************************************************************************/
static void
RecordId( char* dest, char* src, int lenSrc )
{
   int	lenDest;

#ifdef HYPERSRC_DEBUG /* too frequent for HYPERSRC_PEDANTIC */
   if ( !dest  ||  !src )
   {
      YACC_INTERNAL_ERROR_MSG( "\n## Invalid args passed to RecordId()." );
      YACC_ABORT();
   }
#endif

   lenDest = MIN( lenSrc, MAX_ID_LEN-1 );
   memcpy( dest, src, lenDest );
   dest[lenDest] = '\0';

   return;
}

/*****************************************************************************
 * Called when a function definition is completely parsed (ie '}' is encountered).
 *****************************************************************************/
static void
Yacc_stmt_func_def( void )
{
   AddFuncDefToGraph( funcDefId, lineLeftParenthesis, lineRightBrace );
   return;


/* -------------- DEPRECATED ---------------- */
#if 0
   switch ( parserMode )
   {
      case PARSER_MODE_HYPERSRC:
         AddFuncDefToGraph( funcDefId, lineLeftParenthesis, lineRightBrace );
         return;

      case PARSER_MODE_PRINTF:
         printf( "* %s{}", funcDefId );
#ifdef DEBUG
         printf( "    (%d..%d)", lineLeftParenthesis, lineRightBrace );
#endif
         printf( "\n\n\n" );
         return;
   }
#endif
/* -------------- DEPRECATED ---------------- */
}

/*****************************************************************************
 * Called when an identifier of a function call is parsed.
 *
 * Parms : pFuncCallIdMaybeExtra
 *         Must point to the first char of a presumed identifier,
 *         but doesn't matter if extra chars past the identifier were lexed.
 *         (Lex scoops more than an id for TOKEN_FUNC_CALL.)
 *
 *         lenFuncCallId
 *         Must be the exact length of the identifier portion.
 *****************************************************************************/
static void
Yacc_func_call_id( char* pFuncCallIdMaybeExtra, /* !! might be more than an id !! */
                   int   lenFuncCallId )        /* length of the identifier portion */
{
  /*
   * Record just the identifier of the function call.
   * funcCallId (as opposed to pFuncCallIdMaybeExtra)
   * will hold just the identifer, nothing more.
   */
   RecordId( funcCallId, pFuncCallIdMaybeExtra, lenFuncCallId );

  /*
   * Graph this function call.
   * Unless what was presumed to be an identifier of a function call
   * is really a C keyword.
   */
   if ( ShouldFuncIdBeGraphed( funcCallId ) )
      AddFuncCallToGraph( funcCallId, lenFuncCallId, lineLex );

   return;


/* -------------- DEPRECATED ---------------- */
#if 0
  /*
   * Graph this function call.
   */
   switch ( parserMode )
   {
      case PARSER_MODE_HYPERSRC:
         AddFuncCallToGraph( funcCallId, lineLex );
         return;

      case PARSER_MODE_PRINTF:
  #if 0
         /* try to resemble ctags style */
         printf( "%s() in %s{}   call  %d   \n", funcCallId, funcDefId, 0 );
  #else
         /* more readable */
         printf( "    +--> %s() called", funcCallId );
    #ifdef DEBUG
         printf( "    (%d)", lineLex );
    #endif /*DEBUG*/
         printf( "\n" );
  #endif
         return;
   }
#endif
/* -------------- DEPRECATED ---------------- */

}

/*****************************************************************************
 * Count [a-Z_] chars.
 *****************************************************************************/
static int
CountIdChars( const char* pc )
{
   int	len;
   char	c;

   for ( len = 0; len <= MAX_ID_LEN; ++len )
   {
      c = pc[len];
      if ( isalnum(c)  ||  c == '_' )
         continue;
      break;
   }

   return len;
}


%}

%%


/*===========================================================================*/


/*****************************************************************************
 * This set of context/tricks/kludges are employed to simplify/reduce the yacc rules
 * which otherwise would have been far more complex and difficult to write.
 *
 * - a parenthesis context: lex really returns nested parentheses  as TEXT tokens
 *                          unparenthesized commas returned as TEXT tokens
 * - nested parentheses   : lex only passes nested parentheses when the parser
 *                          is parsing the parameter list of a function call.
 * - a curly brace context: lex really returns nested curly braces as TEXT tokens
 * - a C comment   context: lex returns everything comment as TEXT tokens
 * - type casts           : lex filters out type casts
 * - "if ("               : lex filters out the parenthesis so that func calls
 *                          within a if() will be parsed (usually nested parentheses
 *                          are ignored).
 * - function prototypes  : filtered out by lex
 *
 * Notes: The "error" token should be placed in a rule that is closest
 *        to where an error might occur, so that error recovery will
 *        occur sooner.  Writing the "error" token in the "program" or "stmt_s"
 *        rules would cause a lot of input to be skipped while recovering.
 *****************************************************************************/

program:
		stmt_s
	|	TOKEN_ERROR
	|	TOKEN_IGNORE
	;

stmt_s:
		stmt
	|	stmt_s   stmt
	;

stmt:
		stmt_func_def
	|   text_id_s   '{'   text_id_s   '}'
	;

/*****************************************************************************
 * function definition rules
 *
 * text_id_s_or_nothing accounts for old C style code in which parm types
 * were written between ')' and '{'.
 *****************************************************************************/

stmt_func_def:
		stmt_func_def_id   stmt_func_def_left_par   stmt_func_def_parm_s_or_nothing   stmt_func_def_right_par   text_id_s_or_nothing   stmt_func_def_left_brace   stmt_func_body_or_nothing   stmt_func_def_right_brace
								{ ParserReset( RESET_STMT ); Yacc_stmt_func_def(); }
	|   error   TOKEN_TEXT		{ ParserErrorRecover(); yyerrok; }
	|   error   TOKEN_ID		{ ParserErrorRecover(); yyerrok; }
	;

stmt_func_def_left_par:
		'('						{ lineLeftParenthesis = lineLex;
                                  parsedLeftParOfFuncDef = 1; }
	;

stmt_func_def_right_par:
		')'						{ betweenParmsAndBody = 1; } /* parser is between list of parms and body of func def */
	;

stmt_func_def_left_brace:
		'{'						{ betweenParmsAndBody = 0;
                                  lineLeftBrace = lineLex; }
	;

stmt_func_def_right_brace:
		'}'						{ lineRightBrace = lineLex; }
	;

stmt_func_def_id:
		text_id_s				{ parsingFuncDef = TRUE; RecordId( funcDefId, parsedId, parsedIdLen ); }
	;

stmt_func_def_parm_s_or_nothing:
		{ /* nothing */ }
	|	stmt_func_def_parm_s
	;

stmt_func_def_parm_s:
		text_id_s
	|	stmt_func_def_parm_s   ','   text_id_s
	;

stmt_func_body_or_nothing:
		{ /* nothing */ }
	|	stmt_func_body
	;

stmt_func_body:
		stmt_func_body_
	|	stmt_func_body   stmt_func_body_
	;

stmt_func_body_:
		text_id
	|	func_call
	|	')'		/* suppress a parse error */
	;

/*****************************************************************************
 * function call rules
 *
 * Notes: Conditional compilation can cause a parse error in the func_call* rules.
 *        Eg:
 *              foo(
 *        #if 0
 *              a, b);
 *        #else
 *              a, b, c);
 *        #endif
 *
 *        "yyvalstr" is a kludge to suppress gcc warning that $<str>$
 *        is uninitialized (warning appears when using gcc's -O3/-inline).
 *****************************************************************************/

func_call:
		func_call_id   func_call_left_par   func_call_parm_s_list_or_nothing   func_call_right_par
	|	TOKEN_FUNC_CALL			{ /* Lexeme (yyvalstr) is an entire function call. */
                                  /* "lexedIdLen" is N/A in this case, so count id chars. */
                                  char* yyvalstr = $<str>$;
                                  Yacc_func_call_id( yyvalstr, CountIdChars(yyvalstr) ); }
	|   error   TOKEN_TEXT		{ yyerrok; }
	|   error   TOKEN_ID		{ yyerrok; }
	;

func_call_id:
		TOKEN_ID				{ char* yyvalstr = $<str>$;
                                     #ifdef HYPERSRC_DEBUG
                                     if ( lexedIdLen != strlen(yyvalstr) )
                                     { YACC_INTERNAL_ERROR_MSG( "lexedIdLen is wrong" ); }
                                     #endif
                                  Yacc_func_call_id( yyvalstr, lexedIdLen ); }
	;

func_call_left_par:
		'('						{ ++parsingCallParms; }
	;

func_call_right_par:
		')'						{ --parsingCallParms; }
	;

func_call_parm_s_list_or_nothing:
		{ /* nothing */ }
	|	func_call_parm_s_list
	;

func_call_parm_s_list:
		func_call_parm_s
	|	func_call_parm_s_list   ','   func_call_parm_s
	;

func_call_parm_s:
		func_call_parm
	|	func_call_parm_s   func_call_parm
	;

func_call_parm:
		func_call
	|	TOKEN_TEXT
	|	TOKEN_ID
	;

/*****************************************************************************
 * misc rules
 *****************************************************************************/

text_id_s_or_nothing:
		{ /* nothing */ }
	|	text_id_s
	;

text_id_s:
		text_id
	|	text_id_s   text_id
	;

text_id:
		TOKEN_TEXT		{/*char* yyvalstr = $<str>$; parsedIdLen = strlen(yyvalstr); RecordId( parsedId, yyvalstr, parsedIdLen )*/; }
	|	TOKEN_ID		{ char* yyvalstr = $<str>$; parsedIdLen = strlen(yyvalstr); RecordId( parsedId, yyvalstr, parsedIdLen ); }
	;

%%


/*===========================================================================*/


/*----------------------------------------------------------------------------
 * public functions
 *---------------------------------------------------------------------------*/


/*****************************************************************************
 * For hypersrc, parse functions in a module (a source file).
 *****************************************************************************/
void
ParseFuncsInModule( char* pInput, unsigned int inputSize, char* pInputName, int debug )
{
   void*	hnd = NULL;

   if ( pInputName )
      strncpy( parserInputName, pInputName, sizeof(parserInputName) );
   else
      parserInputName[0] = 0;

#ifdef HYPERSRC_PEDANTIC
   if ( pInput == NULL ) { YACC_INTERNAL_ERROR_MSG( "ParseFuncsInModule() pInput==NULL." ); YACC_ABORT(); }
#endif

   ParserCreate( debug ? PARSER_MODE_PRINTF : PARSER_MODE_HYPERSRC,
                 inputSize * 2 /* extra to be safe */ );
   hnd = LexBufferUse( pInput );
      yyparse();
   LexBufferDiscard( hnd );
   ParserDelete();

   return;
}


/*----------------------------------------------------------------------------
 * internal functions
 *---------------------------------------------------------------------------*/


/*****************************************************************************
 * Initialize the parser.
 *
 * Parms:		mode
 *				The parser's mode: one of PARSER_MODE_*.
 *
 *				strHeapSize
 *				The size of string heap to be allocated.
 *              A safe size would be double the size of the input to be parsed.
 *****************************************************************************/
static void
ParserCreate( int mode, int strHeapSize )
{
#ifdef HYPERSRC_PEDANTIC
   if ( initialized ) { YACC_INTERNAL_ERROR_MSG( "ParserCreate() can't be nested." ); YACC_ABORT(); }
#endif

  /*
   * Reset state.
   */
   ParserReset( RESET_FULL );

  /*
   * Record parser's mode.
   */
   parserMode = mode;

  /*
   * Create a heap to hold strings that will be parsed.
   */
   StrHeapCreate( &strHeap, strHeapSize );

   return;
}

/*****************************************************************************
 * Delete resources used by parser.
 *****************************************************************************/
static void
ParserDelete( void )
{
   initialized = FALSE;

   StrHeapDelete( &strHeap );
}

/*****************************************************************************
 * Reset parser.
 * Pass RESET_*.
 *****************************************************************************/
static void
ParserReset( int howMuchToReset )
{
   switch ( howMuchToReset )
   {
      case RESET_FULL:
      initialized			= TRUE;
      lineLeftParenthesis	= 0;
      lineLeftBrace			= 0;
      lineRightBrace		= 0;

      /* -- fall thru -- */

      case RESET_STMT: case RESET_ERROR: default:
      parsingFuncDef		= FALSE;
      parsedLeftParOfFuncDef = FALSE;
      parsingCallParms		= FALSE;
      betweenParmsAndBody	= FALSE;
   }
}

/*****************************************************************************
 * Recover from a parse error.
 *****************************************************************************/
static void
ParserErrorRecover( void )
{
   ParserReset( RESET_ERROR );
}

/*****************************************************************************
 * What to do if a parse error is detected in the input.
 *****************************************************************************/
int
yyerror( char* s )
{
   PARSE_ERROR_MSG( "" );
   return 0; /* shutup compiler */
}

/*****************************************************************************
 * For testing.
 *****************************************************************************/
#if 0
#ifdef DEBUG
static char*
TestCase( void )
{
   char* pInput = (char*)malloc( 256 );

   strcpy( pInput, " void foo( int a ) { xxxx } \n" );
   strcpy( pInput, " unsigned int foo( volatile int a , signed char c ) {  ff();  } \n" );
   strcpy( pInput, " int foo( int a ) {  f(); fab ( a, b );  } \n" );
   strcpy( pInput, " int foo( int a )\n{\n\tf(); foo( a, b ); bar(); *p(); pf->virtual();  } \n" );
   strcpy( pInput, " int foo( int a )\n{\n   bar();\n}\nvoid foo2( char c)\n{\n\tbar2( 1 );\n}\n" );
   strcpy( pInput, "int dummy( int a )\n{\n   f();\n}" );

   return pInput;
}
#endif /* DEBUG */
#endif
